/**************************************************************************
 *
 * text_t.h -- a simple 16-bit charater string class
 * Copyright (c) 1998 -- Rodger McNab
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: text_t.h 4 1998-11-17 01:10:39Z sjboddie $
 *
 **************************************************************************/



#ifndef TEXT_T_H
#define TEXT_T_H

#ifndef USE_OBJECTSPACE
#  include <vector>
#  include <list>
#  include <set>
#else
#  include <ospace\std\vector>
#  include <ospace\std\list>
#  include <ospace\std\set>
#endif

// use the standard namespace
#ifdef __GNUG__
// namespaces are broken in current version of gcc
#else
#  ifndef USE_OBJECTSPACE
using namespace std;
#  else
#    ifndef OS_NO_NAMESPACE
using namespace ospace::std;
#    endif
#  endif
#endif


// class prototypes
class inconvertclass;
class outconvertclass;


typedef vector<unsigned short> usvector;

// The class text_t can handle long strings which may contain
// null characters. It uses unsigned shorts to represent up to
// 64K character values.
class text_t {
  public:
  //type support for ucvector
  typedef usvector::iterator iterator;
  typedef usvector::const_iterator const_iterator;
  typedef usvector::reference reference;
  typedef usvector::const_reference const_reference;
  typedef usvector::size_type size_type;
  typedef usvector::difference_type difference_type;
  typedef usvector::const_reverse_iterator const_reverse_iterator;
  typedef usvector::reverse_iterator reverse_iterator;
  
protected:
  usvector text;
  
public:
  // constructors
  text_t ();
  text_t (int i);
  text_t (char *s); // assumed to be a normal c string

  // basic container support
  iterator begin () {return text.begin();}
  const_iterator begin () const {return text.begin();}
  iterator end () {return text.end();}
  const_iterator end () const {return text.end();}

  void erase(iterator pos) {text.erase(pos);}
  void erase(iterator first, iterator last) {text.erase(first, last);}
  void push_back(unsigned short c) {text.push_back(c);}
  void pop_back() {text.pop_back();}
  text_t &operator=(const text_t &x) {text=x.text;return *this;}
  reference operator[](size_type n) {return text[n];};
  const_reference operator[](size_type n) const {return text[n];};

  bool empty () const {return text.empty();}
  size_type size() const {return text.size();}
  friend bool operator!=(const text_t& x, const text_t& y);
  friend bool operator==(const text_t& x, const text_t& y);
  friend bool operator<(const text_t& x, const text_t& y);
  friend bool operator>(const text_t& x, const text_t& y);
  
  // added functionality
  void clear () {text.erase(text.begin(),text.end());}
  void append (const text_t &t);
  void appendrange (iterator first, iterator last);
  void appendrange (const_iterator first, const_iterator last);
  text_t &operator+= (const text_t &t) {append(t);return *this;}

  // support for integers
  void appendint (int i);
  void setint (int i) {clear();appendint(i);}
  text_t &operator=(int i) {setint (i);return *this;}
  text_t &operator+= (int i) {appendint(i);return *this;}
  int getint ();

  // support for arrays of chars
  void appendcarr (char *s, size_type len);
  void setcarr (char *s, size_type len) {clear();appendcarr(s,len);}

  // support for null-terminated C strings
  void appendcstr (char *s);
  void setcstr (char *s) {clear();appendcstr(s);}
  text_t &operator= (char *s) {setcstr(s);return *this;} // c string
  text_t &operator+= (char *s) {appendcstr(s);return *this;} // c string

  // strings returned from getcarr and getcstr become the callers 
  // responsibility and should be deallocated with "delete"
  char *getcarr(size_type &len) const;
  char *getcstr() const;
};


inline text_t operator+(const text_t &t1, const text_t &t2)
{
  text_t tnew = t1;
  tnew.append(t2);
  return tnew;
}

inline text_t operator+(const text_t &t1, int i1)
{
  text_t tnew = t1;
  tnew.appendint(i1);
  return tnew;
}

inline text_t operator+(const text_t &t1, char *s1)
{
  text_t tnew = t1;
  tnew.appendcstr(s1);
  return tnew;
}


inline bool operator!=(const text_t& x, const text_t& y) 
{
  return (x.text != y.text);
}

inline bool operator==(const text_t& x, const text_t& y) 
{
  return (x.text == y.text);
}

inline bool operator<(const text_t& x, const text_t& y) 
{
  return (x.text < y.text);
}

inline bool operator>(const text_t& x, const text_t& y) 
{
  return (x.text > y.text);
}


struct eqtext_t 
{
  bool operator()(const text_t t1, const text_t t2) const
  { return t1 == t2; }
};

struct lttext_t 
{
  bool operator()(const text_t t1, const text_t t2) const
  { return t1 < t2; }
};


// frequently used derived types
typedef set<text_t,lttext_t> text_tset;
typedef list<text_t> text_tlist;  // more efficient for insertions/deletions
typedef vector<text_t> text_tarray; // more space efficient than text_tlist


// general functions which work on text_ts

// find a character within a range
text_t::const_iterator findchar (text_t::const_iterator first, text_t::const_iterator last, 
				 unsigned short c);
text_t::iterator findchar (text_t::iterator first, text_t::iterator last, 
			   unsigned short c);

// get a string up to the next delimiter (which is skipped)
text_t::const_iterator getdelimitstr (text_t::const_iterator first, 
				      text_t::const_iterator last,
				      unsigned short c, text_t &outstr);
text_t::iterator getdelimitstr (text_t::iterator first, text_t::iterator last,
				unsigned short c, text_t &outstr);

// split a string with a character
void splitchar (text_t::const_iterator first, text_t::const_iterator last,
		unsigned short c, text_tlist &outlist);
void splitchar (text_t::const_iterator first, text_t::const_iterator last,
		unsigned short c, text_tarray &outlist);

// join a string using a character
void joinchar (const text_tlist &inlist, unsigned short c, text_t &outtext);
void joinchar (const text_tarray &inlist, unsigned short c, text_t &outtext);

// count the occurances of a character within a range
int countchar (text_t::const_iterator first, text_t::const_iterator last,
	       unsigned short c);








// conversion classes used for getting information in to and out of
// the text_t class.

class convertclass 
{
public:
  enum status_t {finished, stopped, unfinished};
  
  convertclass ();
  virtual void reset ();
};



// convert from a char stream to the text_t class
// the default version assumes the input is a ascii
// character array
class inconvertclass : public convertclass
{
public:
  inconvertclass ();
  void reset ();
  void setinput (char *thestart, size_t thelen);
  virtual void convert (text_t &output, status_t &status);

protected:
  char *start;
  size_t len;
};

// to get something which will do the conversion
// to ascii declare a (non global!) instance like
// this
// inconvertclass ascii2text_t;


// Convert from a text_t class to a char stream
// This default version assumes the output is a ascii
// character array. If you set the output stream you
// can use this class to output to a stream using the
// << operator. The << operator can also be conveniently
// used to set the output stream by doing something like
//
// cout << text_t2ascii << textstr << anothertextstr;
//
// this class assumes that the input text doesn't change
// while the conversion takes place
class outconvertclass : public convertclass
{
public:
  outconvertclass ();
  void reset ();
  void setinput (text_t *theinput);
  // note that convert does not null-terminate the
  // output array of characters
  virtual void convert (char *output, size_t maxlen, 
			size_t &len, status_t &status);

  void setostream (ostream *theouts);
  ostream *getostream ();

protected:
  text_t *input;
  text_t::iterator texthere; // only valid if input is valid

  ostream *outs;
};

// to get something which will do the conversion
// to text_t declare a (non global!) instance like
// this
// outconvertclass text_t2ascii;


// stream operators for the output class
outconvertclass &operator<< (ostream &theouts, outconvertclass &outconverter);
outconvertclass &operator<< (outconvertclass &outconverter, const text_t &t);


#endif
